var
  _ = require('underscore'),
  EventEmitter = require('events').EventEmitter,
  Scope = require('./lib/scope').Scope,
  RouteHandler = require('./lib/routeHandler').RouteHandler,
  DispatchHandler = require('./lib/dispatchHandler').DispatchHandler;

/**
 * Application
 */
function Application(nameToStores){

  this._stores = _.values(nameToStores);

  EventEmitter.call(this);
  _.extend(this, EventEmitter.prototype);

  Scope.call(this, nameToStores);
  this.get = Scope.prototype.get;

  RouteHandler.call(this, this._stores);
  _.extend(this, RouteHandler.prototype);

  DispatchHandler.call(this, this._stores);
  this.dispatch = _.bind(function(event, payload){
    return DispatchHandler.prototype.dispatch.call(this, event, payload);
  }, this);
  this.emitChange = _.bind(function(){
    return Application.prototype.emitChange.call(this);
  }, this);

  var self = this;
  _.each(this._stores, function(store){
    //when a store fires a change event, we forward to the root view listening
    store.on('change', function(){
      self.emitChange(store);
    });
    store.go = _.bind(self.go, self);
    store.scope = _.bind(self.push, self);
  });
}

Application.prototype.push = function(map){
  var result = Scope.prototype.push.call(this, map);
  result.dispatch = this.dispatch;
  result.emitChange = this.emitChange;
  result.push = Application.prototype.push;
  return result;
};

Application.prototype.emitChange = function(){
  this.emit('change');
};

/**
 * ViewMixin
 */
var ViewMixin = {
  scope: function(values){
    return values ? this.props.scope.push(values) : this.props.scope;
  },
  resolve: function(key){
    return this.props.scope.get(key);
  },
  dispatch: function(event, payload){
    return this.props.scope.dispatch(event, payload);
  },
  dispatcher: function(event){
    return _.bind(function(e){
      this.props.scope.dispatch(event, e.target.value);
    }, this);
  }
};

/**
 * RootViewMixin
 */
var RootViewMixin = {
  debounceTime: 10,
  componentWillMount: function(){
    this.callback = (function(){
      this.forceUpdate();
    }).bind(this);
    this.props.scope.on('change', _.debounce(this.callback, this.debounceTime));
  }
};

_.extend(RootViewMixin, ViewMixin);

exports.Application = Application;
exports.ViewMixin = ViewMixin;
exports.RootViewMixin = RootViewMixin;